/******************************************************************************
 * $Id: ogr_sxflayer.cpp  $
 *
 * Project:  SXF Translator
 * Purpose:  Definition of classes for OGR SXF Layers.
 * Author:   Ben Ahmed Daho Ali, bidandou(at)yahoo(dot)fr
 *
 ******************************************************************************
 * Copyright (c) 2011, Ben Ahmed Daho Ali
 *
 * Permission is hereby granted, free of charge, to any person obtaining a
 * copy of this software and associated documentation files (the "Software"),
 * to deal in the Software without restriction, including without limitation
 * the rights to use, copy, modify, merge, publish, distribute, sublicense,
 * and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included
 * in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
 * OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 * DEALINGS IN THE SOFTWARE.
 ****************************************************************************/
#define  _USE_MATH_DEFINES

#include "ogr_sxf.h"
#include "cpl_conv.h"
#include "cpl_string.h"
#include "ogr_p.h"
#include "ogr_srs_api.h"

CPL_CVSID("$Id: ogrsxflayer.cpp $");

/************************************************************************/
/*                        OGRSXFLayer()                                 */
/************************************************************************/

OGRSXFLayer::OGRSXFLayer( VSILFILE* fp, const char* pszLayerName , 
                                    RecordSXFPSP  oSXFP, RecordSXFDSC  oSXFD )
{
    fpSXF = fp;
    oSXFPSP = oSXFP;
    oSXFDSC = oSXFD;
    nNextFID = 0;
    bIncorrectFType = FALSE;
    bEOF = FALSE;

    poSRS = new OGRSpatialReference(SRS_WKT_WGS84);  
    
/****************************************************
 * TODO : adapte the code below

    double ScaleFactor = 1.0, ZoneNumber = 30; // to improve 

    double padfPrjParams[8] = {oSXFP.dfMainPar1, oSXFP.dfMainPar2, oSXFP.dfMainPtPar, 
                               oSXFP.dfAxialMer,      ScaleFactor,  
                               oSXFP.dfFalseEasting, oSXFP.dfFalseNorthing, ZoneNumber };

    poSRS->importFromPanorama( oSXFP.iProjSys, oSXFP.iDatum, oSXFP.iEllips, padfPrjParams );

*/

    poFeatureDefn = new OGRFeatureDefn( pszLayerName );
    poFeatureDefn->Reference();

    OGRFieldDefn oCodeField = OGRFieldDefn( "OBJECTCODE", OFTInteger );
    oCodeField.SetWidth(10);
    poFeatureDefn->AddFieldDefn( &oCodeField );

    OGRFieldDefn oNumField = OGRFieldDefn( "OBJECTNUMB", OFTInteger );
    oNumField.SetWidth(10);
    poFeatureDefn->AddFieldDefn( &oNumField );

    OGRFieldDefn  oTextField( "TEXT", OFTString );
    oTextField.SetWidth(255);
    poFeatureDefn->AddFieldDefn( &oTextField );

//---------------------------------------------------------------------
// TODO : adding other sxf standards fields 
//---------------------------------------------------------------------

}

/************************************************************************/
/*                         ~OGRSXFLayer()                               */
/************************************************************************/

OGRSXFLayer::~OGRSXFLayer()

{
    if( poSRS != NULL )
        poSRS->Release();

    poFeatureDefn->Release();

 }


/************************************************************************/
/*                            ResetReading()                            */
/************************************************************************/

void OGRSXFLayer::ResetReading()

{
    nNextFID = 0;
    bIncorrectFType = FALSE;
    bEOF = FALSE;
    VSIFSeekL( fpSXF, sizeof (RecordSXFHEAD)+ sizeof (RecordSXFPSP)+ 
                                              sizeof (RecordSXFDSC), SEEK_SET );
}


/************************************************************************/
/*                           GetNextFeature()                           */
/************************************************************************/

OGRFeature *OGRSXFLayer::GetNextFeature()
{
           OGRFeature  *poFeature;

    while(TRUE)
    {
        if (bEOF)
            return NULL;


        poFeature = GetNextRawFeature();
        
        if ( bIncorrectFType == TRUE) 
        {         
            bIncorrectFType = FALSE ;
            continue;
        }

        if (poFeature == NULL)
        return NULL;
    
        if((m_poFilterGeom == NULL
            || FilterGeometry( poFeature->GetGeometryRef() ) )
        && (m_poAttrQuery == NULL
            || m_poAttrQuery->Evaluate( poFeature )) )
        {
            return poFeature;
        }
        else
            delete poFeature;
    }    
  
}

/************************************************************************/
/*                           TestCapability()                           */
/************************************************************************/

int OGRSXFLayer::TestCapability( const char * pszCap )

{
    return FALSE;
}

/************************************************************************/
/*                                TranslateXYH()                        */
/************************************************************************/
/****
 * TODO : Take into account informations given in the passport 
 * like unit of mesurement, type and dimensions (integer, float, double) of coordinate,
 * the vector format ....
 */

GUInt32 OGRSXFLayer::TranslateXYH ( RecordSXFOBJ  oSXFObj,     char *psBuff, 
                          long double *dfX, long double *dfY, double *dfH)
{
    *dfY = *(long double *)(psBuff);
    *dfX = *(long double *)(psBuff + 8);

    if (dfH != NULL)    
        *dfH = *(double *)(psBuff + 16); 

    if ( oSXFPSP.iPlanUoM == 64) // radians  - to enhance
    {
        *dfX  *= (180.0/M_PI);   
        *dfY  *= (180.0/M_PI);     
    }

// dimension in byte of coordiantes buffer
    if (dfH != NULL)    return 16+4; 
    else                return 16;
} // GetXYH()

/************************************************************************/
/*                       OGRSXFPointLayer()                             */ //-------------------------------------
/************************************************************************/

OGRSXFPointLayer::OGRSXFPointLayer( VSILFILE* fp, const char* pszLayerName, 
                                            RecordSXFPSP  oSXFP, RecordSXFDSC  oSXFD ) :
                                            OGRSXFLayer(fp, pszLayerName, oSXFP, oSXFD )
{
    poFeatureDefn->SetGeomType( wkbMultiPoint );

}
/************************************************************************/
/*                         GetNextRawFeature()                          */
/************************************************************************/

OGRFeature *OGRSXFPointLayer::GetNextRawFeature()
{
    OGRFeature *poFeature = NULL;
    RecordSXFOBJ  oSXFObj;

        if ( VSIFReadL( &oSXFObj, sizeof(RecordSXFOBJ), 1, fpSXF ) < 1 
            || nNextFID == oSXFDSC.nObjCount )
        {
            bEOF = TRUE;
            return NULL;
        }

        size_t nDataSize = oSXFObj.nRecordLength - sizeof(RecordSXFOBJ);

        char * psRecordBuf = (char *) CPLMalloc( nDataSize );

        VSIFReadL( psRecordBuf, nDataSize , 1, fpSXF );

        if (oSXFObj.bGeomType == sxfPoint) 
            poFeature = TranslatePoint( oSXFObj, psRecordBuf );

        if (poFeature != NULL)
        {
        poFeature->SetField("OBJECTCODE", oSXFObj.iCC);
        poFeature->SetField("OBJECTNUMB", oSXFObj.nObjNumb);
        poFeature->SetFID( nNextFID++ );
        }
        else     bIncorrectFType = TRUE ;
    
        CPLFree( psRecordBuf );

    return poFeature;
}

/************************************************************************/
/*                         TranslatePoint   ()                          */
/************************************************************************/
OGRFeature *OGRSXFPointLayer::TranslatePoint( RecordSXFOBJ  oSXFObj, char * psRecordBuf )
{
        long double dfX = 1.0;
        long double dfY = 1.0;
        GUInt32 nOffset = 0;

        OGRFeature *poFeature = new OGRFeature( poFeatureDefn );
        OGRMultiPoint* poMPt = new OGRMultiPoint();

        nOffset += TranslateXYH( oSXFObj, psRecordBuf , &dfX, &dfY ) ;

        poMPt->addGeometryDirectly( new OGRPoint( dfX, dfY ) );

/*---------------------- Reading SubObjects --------------------------------*/

    for(int count=0 ; count <  oSXFObj.nSubObjCount ; count++)
    {
    char * psBuff = psRecordBuf + nOffset;

    GUInt16 nSubObj = *(GUInt16*) psBuff; 
    GUInt16 nCoords = *(GUInt16*) (psBuff + 2); 


    nOffset +=4;

        for (int i=0; i < nCoords ; i++)
        {
        char * psCoords = psRecordBuf + nOffset ;

        nOffset +=  TranslateXYH( oSXFObj, psCoords , &dfX, &dfY ) ;

        poMPt->addGeometryDirectly( new OGRPoint( dfX, dfY ) );
        } 
    } // for count

/*****
 * TODO : 
 *          - Translate graphics 
 *          - Translate 3D vector
 */

        poFeature->SetGeometryDirectly( poMPt );

        return poFeature;
}

/************************************************************************/
/*                       OGRSXFLineLayer()                              */   //------------------------------------------
/************************************************************************/

OGRSXFLineLayer::OGRSXFLineLayer( VSILFILE* fp, const char* pszLayerName,  
                                            RecordSXFPSP  oSXFP, RecordSXFDSC  oSXFD ) :
                                            OGRSXFLayer(fp, pszLayerName, oSXFP, oSXFD )
{
    poFeatureDefn->SetGeomType( wkbMultiLineString );

}
/************************************************************************/
/*                         GetNextRawFeature()                          */
/************************************************************************/

OGRFeature *OGRSXFLineLayer::GetNextRawFeature()
{
    OGRFeature *poFeature = NULL;
    RecordSXFOBJ  oSXFObj;


        if ( VSIFReadL( &oSXFObj, sizeof(RecordSXFOBJ), 1, fpSXF ) < 1 
            || nNextFID == oSXFDSC.nObjCount )
        {
            bEOF = TRUE;
            return NULL;
        }

        size_t nDataSize = oSXFObj.nRecordLength - sizeof(RecordSXFOBJ);

        char * psRecordBuf = (char *) CPLMalloc( nDataSize );

        VSIFReadL( psRecordBuf, nDataSize , 1, fpSXF );

        if (oSXFObj.bGeomType == sxfLine )       
            poFeature = TranslateLine ( oSXFObj, psRecordBuf );     

        if (poFeature != NULL)
        {
        poFeature->SetField("OBJECTCODE", oSXFObj.iCC);
        poFeature->SetField("OBJECTNUMB", oSXFObj.nObjNumb);
        poFeature->SetFID( nNextFID++ );
        }
        else     bIncorrectFType = TRUE ;


        CPLFree( psRecordBuf );

        return poFeature;
}
/************************************************************************/
/*                         TranslateLine    ()                          */
/************************************************************************/

OGRFeature *OGRSXFLineLayer::TranslateLine( RecordSXFOBJ  oSXFObj, char * psRecordBuf )
{
        long double dfX = 1.0;
        long double dfY = 1.0;
        GUInt32 nOffset = 0;

        OGRFeature *poFeature = new OGRFeature( poFeatureDefn );
        OGRMultiLineString *poMLS = new  OGRMultiLineString ();

/*---------------------- Reading Primary Line --------------------------------*/

    OGRLineString* poLS = new OGRLineString();

    for(int count=0 ; count <  oSXFObj.nPointsCount ; count++)
    {
        char * psCoords = psRecordBuf + nOffset ;

        nOffset += TranslateXYH( oSXFObj, psCoords , &dfX, &dfY ) ;

        poLS->addPoint( dfX  , dfY );

    }    // for

    poMLS->addGeometry( poLS );

/*---------------------- Reading Sub Lines --------------------------------*/

    for(int count=0 ; count <  oSXFObj.nSubObjCount ; count++)
    {
    poLS->empty();
    char * psBuff = psRecordBuf + nOffset;

    GUInt16 nSubObj = *(GUInt16*) psBuff; 
    GUInt16 nCoords = *(GUInt16*) (psBuff + 2); 

    nOffset +=4;

        for (int i=0; i < nCoords ; i++)
        {
        char * psCoords = psRecordBuf + nOffset ;

         nOffset +=  TranslateXYH( oSXFObj, psCoords , &dfX, &dfY ) ;

        poLS->addPoint( dfX  , dfY );
        }

    poMLS->addGeometry( poLS );
    }    // for

        delete poLS;
        poFeature->SetGeometryDirectly( poMLS );

/*****
 * TODO : 
 *          - Translate graphics 
 *          - Translate 3D vector
 */

        return poFeature;
}
/************************************************************************/
/*                       OGRSXFPolygonLayer()                           */   //----------------------------------------------
/************************************************************************/

OGRSXFPolygonLayer::OGRSXFPolygonLayer( VSILFILE* fp, const char* pszLayerName,  
                                            RecordSXFPSP  oSXFP, RecordSXFDSC  oSXFD ) :
                                            OGRSXFLayer(fp, pszLayerName, oSXFP, oSXFD )
{
    poFeatureDefn->SetGeomType( wkbPolygon );

}
                                            
/************************************************************************/
/*                         GetNextRawFeature()                          */
/************************************************************************/

OGRFeature *OGRSXFPolygonLayer::GetNextRawFeature()
{
    OGRFeature *poFeature = NULL;
    RecordSXFOBJ  oSXFObj;

        if ( VSIFReadL( &oSXFObj, sizeof(RecordSXFOBJ), 1, fpSXF ) < 1 
            || nNextFID == oSXFDSC.nObjCount )
        {
            bEOF = TRUE;
            return NULL;
        }

        size_t nDataSize = oSXFObj.nRecordLength - sizeof(RecordSXFOBJ);

        char * psRecordBuf = (char *) CPLMalloc( nDataSize );

        VSIFReadL( psRecordBuf, nDataSize , 1, fpSXF );

        if (oSXFObj.bGeomType == sxfPolygon ) 
            poFeature = TranslatePolygon( oSXFObj, psRecordBuf );

        if (poFeature != NULL)
        {
        poFeature->SetField("OBJECTCODE", oSXFObj.iCC);
        poFeature->SetField("OBJECTNUMB", oSXFObj.nObjNumb);
        poFeature->SetFID( nNextFID++ );
        }
        else     bIncorrectFType = TRUE ;


        CPLFree( psRecordBuf );

        return poFeature;
}


/************************************************************************/
/*                         TranslatePolyg   ()                          */
/************************************************************************/
OGRFeature *OGRSXFPolygonLayer::TranslatePolygon( RecordSXFOBJ  oSXFObj, char * psRecordBuf )
{
    long double dfX = 1.0;
    long double dfY = 1.0;
    GUInt32 nOffset = 0;

        OGRFeature *poFeature = new OGRFeature( poFeatureDefn );
        OGRPolygon *poPoly = new OGRPolygon();
        OGRLineString* poLS = new OGRLineString();

/*---------------------- Reading Primary Polygon --------------------------------*/

    for(int count=0 ; count <  oSXFObj.nPointsCount ; count++)
    {
        char * psBuf = psRecordBuf + nOffset ;

        nOffset +=  TranslateXYH( oSXFObj, psBuf , &dfX, &dfY ) ;

        poLS->addPoint( dfX  , dfY );

    }    // for

        OGRLinearRing *poLR = new OGRLinearRing();
        poLR->addSubLineString( poLS, 0 );

        poPoly->addRingDirectly( poLR );

/*---------------------- Reading Sub Lines --------------------------------*/

    for(int count=0 ; count <  oSXFObj.nSubObjCount ; count++)
    {
    poLS->empty();
    char * psBuff = psRecordBuf + nOffset;

    GUInt16 nSubObj = *(GUInt16*) psBuff; 
    GUInt16 nCoords = *(GUInt16*) (psBuff + 2); 

    nOffset +=4;

        for (int i=0; i < nCoords ; i++)
        {
        char * psCoords = psRecordBuf + nOffset ;

         nOffset +=  TranslateXYH( oSXFObj, psCoords , &dfX, &dfY ) ;

        poLS->addPoint( dfX  , dfY );
        }

        OGRLinearRing *poLR = new OGRLinearRing();
        poLR->addSubLineString( poLS, 0 );

        poPoly->addRingDirectly( poLR );
    }    // for

        poFeature->SetGeometryDirectly( poPoly );   //poLS);
        delete poLS;

/*****
 * TODO : 
 *          - Translate graphics 
 *          - Translate 3D vector
 */
        return poFeature;
}

/************************************************************************/
/*                       OGRSXFTextLayer()                              */ //-----------------
/************************************************************************/

OGRSXFTextLayer::OGRSXFTextLayer( VSILFILE* fp, const char* pszLayerName,  
                                              RecordSXFPSP  oSXFP, RecordSXFDSC  oSXFD ) :
                                            OGRSXFLayer(fp, pszLayerName, oSXFP, oSXFD )
{
    poFeatureDefn->SetGeomType( wkbLineString );

}
/************************************************************************/
/*                         GetNextRawFeature()                          */
/************************************************************************/

OGRFeature *OGRSXFTextLayer::GetNextRawFeature()
{
    OGRFeature *poFeature = NULL;
    RecordSXFOBJ  oSXFObj;


        if ( VSIFReadL( &oSXFObj, sizeof(RecordSXFOBJ), 1, fpSXF ) < 1 
            || nNextFID == oSXFDSC.nObjCount )
        {
            bEOF = TRUE;
            return NULL;
        }

        size_t nDataSize = oSXFObj.nRecordLength - sizeof(RecordSXFOBJ);

        char * psRecordBuf = (char *) CPLMalloc( nDataSize );

        VSIFReadL( psRecordBuf, nDataSize , 1, fpSXF );

        if (oSXFObj.bGeomType == sxfText ) 
            poFeature = TranslateText ( oSXFObj, psRecordBuf );

        if (poFeature != NULL)
        {
        poFeature->SetField("OBJECTCODE", oSXFObj.iCC);
        poFeature->SetField("OBJECTNUMB", oSXFObj.nObjNumb);
        poFeature->SetFID( nNextFID++ );
        }
        else     bIncorrectFType = TRUE ;


        CPLFree( psRecordBuf );

        return poFeature;
}
/************************************************************************/
/*                         TranslateText    ()                          */
/************************************************************************/
OGRFeature *OGRSXFTextLayer::TranslateText( RecordSXFOBJ  oSXFObj, char * psRecordBuf )
{
    long double dfX = 1.0;
    long double dfY = 1.0;
    GUInt32 nOffset = 0;

        OGRFeature *poFeature = new OGRFeature( poFeatureDefn );
        OGRLineString* poLS = new OGRLineString();

    for(int count=0 ; count <  oSXFObj.nPointsCount ; count++)
    {
        char * psBuf = psRecordBuf + nOffset;

        nOffset += TranslateXYH( oSXFObj, psBuf , &dfX, &dfY ) ;

        poLS->addPoint( dfX  , dfY );

    }    // for


        poFeature->SetGeometryDirectly( poLS );

/*------------------     READING TEXT VALUE   ---------------------------------------*/

if ( oSXFObj.nSubObjCount == 0)
  {
      char * pszTxt = psRecordBuf +(16*oSXFObj.nPointsCount);

        GByte nTextL = (GByte) *pszTxt;

        char * pszTextBuf = (char *)CPLMalloc( nTextL+1 );

        strncpy(pszTextBuf, (pszTxt+1),    nTextL+1);

        poFeature->SetField( "TEXT", pszTextBuf );
 
//         poFeature->SetStyleString("LABEL(f:\"Arial\",t:\"Hello ...\",s:\"20\")");
//       printf("\n Text lenght = %#d text : %s ",nTextL, pszTextBuf);

        CPLFree( pszTextBuf );  
  }


/*****
 * TODO : 
 *          - Translate graphics 
 *          - Translate 3D vector
 */

        return poFeature;
}





